class FieldType():
    AGGREGATE = "AGGREGATE"  # Returns an aggregate
    FUNCTION = "FUNCTION"    # Returns a function over a field
    FIELD = "FIELD"          # Returns a field from the tuple
    LITERAL = "LITERAL"      # Returns a literal (string, int, or float)
    UNDEFINED = "UNDEFINED"  # Not a legitimate field---shouldn't appear in parsed query

class ReturnType():
    INTEGER = "INTEGER"
    FLOAT = "FLOAT"
    STRING = "STRING"
    DATETIME = "DATETIME"
    UNDEFINED = "UNDEFINED"  # Not a legitimate type---shouldn't appear in parsed query

class FieldDescriptor():
    def __init__(self, alias, underlying_fields, field_type, return_type, aggregate_factory=None, func_factory=None, literal_value=None):
        self.alias = alias
        self.underlying_fields = underlying_fields
        self.field_type = field_type
        self.return_type = return_type
        self.aggregate_factory = aggregate_factory
        self.func_factory = func_factory
        self.function = None
        if func_factory != None:
            self.function = func_factory()
        self.literal_value = literal_value
        self.visible = True
    def __eq__(self, other):
        if isinstance(other, FieldDescriptor):
            return (self.alias == other.alias) and \
               (self.underlying_fields == other.underlying_fields) and \
               (self.field_type == other.field_type) and \
               (self.return_type == other.return_type) and \
               (self.aggregate_factory == other.aggregate_factory) and \
               (self.func_factory == other.func_factory) and \
               (self.literal_value == other.literal_value)
        else:
            return NotImplemented
    def __ne__(self, other):
        result = self.__eq__(other)
        if result is NotImplemented:
            return result
        else:
            return not result



from aggregation import Aggregator
from field_descriptor import FieldDescriptor
from query import QueryTokens
from twitter_fields import TwitterFields

class StatusSource(object):
    TWITTER_FILTER = 1
    TWITTER_SAMPLE = 2

class QueryOperator(object):
    """
        QueryOperator represents an operator in the stream query plan.
        Anyone who implements this class should implement the filter method,
        which takes a list of updates on the stream and returns the ones
        which match the filter.
    """
    def __init__(self):
        pass
    def filter(self, updates, return_passes, return_fails):
        raise  NotImplementedError()
    def filter_params(self):
        """
            Returns a tuple with lists: (follow_ids, track_words)
        """
        raise  NotImplementedError()
    def assign_descriptor(self, tuple_descriptor):
        raise NotImplementedError()
    def can_query_stream(self):
        return False
    def get_tuple_descriptor(self):
        return self.tuple_descriptor

class AllowAll(QueryOperator):
    """
        Allows all updates to pass this filter
    """
    def __init__(self):
        QueryOperator.__init__(self)
    def filter(self, updates, return_passes, return_fails):
        passes = updates
        for update in passes:
            update.set_tuple_descriptor(self.tuple_descriptor)
        fails = []
        if not return_passes:
            passes = None
        if not return_fails:
            fails = None
        return (passes, fails)
    def assign_descriptor(self, tuple_descriptor):
        self.tuple_descriptor = tuple_descriptor

class And(QueryOperator):
    """
        Logical AND between the operators that make up its children.
    """
    def __init__(self, children):
        QueryOperator.__init__(self)
        self.children = children
        self.can_query_stream_cache = self.can_query_stream_impl()
    def filter(self, updates, return_passes, return_fails):
        passes = updates
        fails = []
        for child in self.children:
            (passes, fails_local) = child.filter(passes, True, return_fails)
            if return_fails:
                fails.extend(fails_local)
        if not return_passes:
            passes = None
        if not return_fails:
            fails = None
        return (passes, fails)
    def filter_params(self):
        for child in self.children:
            if child.can_query_stream():
                return child.filter_params()
    def can_query_stream(self):
        return self.can_query_stream_cache
    def can_query_stream_impl(self):
        for child in self.children:
            if child.can_query_stream():
                return True
        return False
    def assign_descriptor(self, tuple_descriptor):
        self.tuple_descriptor = tuple_descriptor
        for child in self.children:
            child.assign_descriptor(tuple_descriptor)

class Or(QueryOperator):
    """
        Logical OR between the operators that make up its children.
    """
    def __init__(self, children):
        QueryOperator.__init__(self)
        self.children = children
        self.can_query_stream_cache = self.can_query_stream_impl()
    def filter(self, updates, return_passes, return_fails):
        passes = []
        fails = updates
        for child in self.children:
            (passes_local, fails) = child.filter(fails, return_passes, True)
            if return_passes:
                passes.extend(passes_local)
        if not return_passes:
            passes = None
        if not return_fails:
            fails = None
        return (passes, fails)
    def filter_params(self):
        (follow_ids, track_words) = ([], []) 
        for child in self.children:
            if child.can_query_stream():
                (follow_local, track_local) = child.filter_params()
                if follow_local != None:
                    follow_ids.extend(follow_local)
                if track_local != None:
                    track_words.extend(track_local)
        return (follow_ids, track_words)
    def can_query_stream(self):
        return self.can_query_stream_cache
    def can_query_stream_impl(self):
        for child in self.children:
            if not child.can_query_stream():
                return False
        return True
    def assign_descriptor(self, tuple_descriptor):
        self.tuple_descriptor = tuple_descriptor
        for child in self.children:
            child.assign_descriptor(tuple_descriptor)

class Not(QueryOperator):
    """
        Logical NOT on the child operator.
    """
    def __init__(self, child):
        QueryOperator.__init__(self)
        self.child = child
    def filter(self, updates, return_passes, return_fails):
        (passes, fails) = self.child.filter(updates, return_fails, return_passes)
        return (fails, passes)
    def filter_params(self):
        return self.child.filter_params()
    def can_query_stream(self):
        return self.child.can_query_stream()
    def assign_descriptor(self, tuple_descriptor):
        self.tuple_descriptor = tuple_descriptor
        self.child.assign_descriptor(tuple_descriptor)

class Follow(QueryOperator):
    """
        Passes updates which contain people in the follower list.
    """
    def __init__(self, ids):
        QueryOperator.__init__(self)
        self.ids = set(ids)
    def filter(self, updates, return_passes, return_fails):
        passes = [] if return_passes else None
        fails = [] if return_fails else None
        for update in updates:
            update.set_tuple_descriptor(self.tuple_descriptor)
            if update.author in self.ids:
                if return_passes:
                    passes.append(update.author)
            elif return_fails:
                fails.append(update.author)
        return (passes, fails)
    def filter_params(self):
        return (self.ids, None)
    def can_query_stream(self):
        return True 
    def assign_descriptor(self, tuple_descriptor):
        self.tuple_descriptor = tuple_descriptor

class Contains(QueryOperator):
    """
        Passes updates which contain the desired term.  Matching is case-insensitive.
    """
    def __init__(self, field_alias, term):
        QueryOperator.__init__(self)
        self.alias = field_alias
        self.term = term.lower()
    def filter(self, updates, return_passes, return_fails):
        passes = [] if return_passes else None
        fails = [] if return_fails else None
        for update in updates:
            update.set_tuple_descriptor(self.tuple_descriptor)
            if self.term in getattr(update, self.alias).lower():
                if return_passes:
                    passes.append(update)
            elif return_fails:
                fails.append(update)
        return (passes, fails)
    def filter_params(self):
        return (None, [self.term])
    def can_query_stream(self):
        if self.alias == TwitterFields.TEXT:
            return True
        else:
            return False
    def assign_descriptor(self, tuple_descriptor):
        self.tuple_descriptor = tuple_descriptor

class Equals(QueryOperator):
    """
        Passes updates which equal the desired term.  Matching is case-sensitive.
    """
    def __init__(self, field_alias, term):
        QueryOperator.__init__(self)
        self.alias = field_alias
        self.term = term
    def filter(self, updates, return_passes, return_fails):
        passes = [] if return_passes else None
        fails = [] if return_fails else None
        for update in updates:
            update.set_tuple_descriptor(self.tuple_descriptor)
            if self.term == getattr(update, self.alias):
                if return_passes:
                    passes.append(update)
            elif return_fails:
                fails.append(update)
        return (passes, fails)
    def filter_params(self):
        return (None, [self.term])
    def can_query_stream(self):
        if self.alias == TwitterFields.TEXT:
            return True
        else:
            return False
    def assign_descriptor(self, tuple_descriptor):
        self.tuple_descriptor = tuple_descriptor

class Location(QueryOperator):
    """
        Passes updates which are located in a geographic bounding box
    """
    def __init__(self, xmin, xmax, ymin, ymax):
        QueryOperator.__init__(self)
        (self.xmin, self.xmax, self.ymin, self.ymax) = (xmin, xmax, ymin, ymax)
    def filter(self, updates, return_passes, return_fails):
        passes = [] if return_passes else None
        fails = [] if return_fails else None
        for update in updates:
            update.set_tuple_descriptor(self.tuple_descriptor)
            x = update.geo.x
            y = update.geo.y
            if x >= xmin and x <= xmax and y >= ymin and y <= ymax:
                if return_passes:
                    passes.append(update)
            elif return_fails:
                fails.append(update)
        return (passes, fails)
    def assign_descriptor(self, tuple_descriptor):
        self.tuple_descriptor = tuple_descriptor

class GroupBy(QueryOperator):
    """
        Groups results from child by some set of fields, and runs the aggregate 
        function(s) over them, emitting results and resetting buckets every
        window seconds.
    """
    def __init__(self, child, groupby, aggregates, window):
        QueryOperator.__init__(self)
        self.child = child
        self.can_query_stream_cache = self.can_query_stream_impl()
        self.groupby = groupby
        self.aggregates = aggregates
        self.window = window
        self.aggregator = Aggregator(self.aggregates, self.groupby, self.window)
    def filter(self, updates, return_passes, return_fails):
        if return_passes:
            (passes, fails) = self.child.filter(updates, return_passes, return_fails)
            new_emissions = []
            new_emissions.extend(self.aggregator.update(passes))
            return (new_emissions, None)
        else:
            return (None, None)
    def filter_params(self):
        return self.child.filter_params()
    def can_query_stream(self):
        return self.can_query_stream_cache
    def can_query_stream_impl(self):
        return self.child.can_query_stream()
    def assign_descriptor(self, tuple_descriptor):
        self.tuple_descriptor = tuple_descriptor
        self.aggregator.tuple_descriptor = tuple_descriptor
        with_aggregates = self.groupby.duplicate()
        for aggregate in self.aggregates:
            with_aggregates.add_descriptor(aggregate)
        with_aggregates.add_descriptor(TwitterFields.created_field)
        for alias, fd in tuple_descriptor.descriptors.items():
            with_aggregates.add_descriptor(fd)
        self.child.assign_descriptor(with_aggregates)



from datetime import timedelta
from datetime import datetime
from tweeql.query import QueryTokens 
from tweeql.exceptions import QueryException
from threading import RLock

class Aggregator():
    def __init__(self, aggregates, groupby, windowsize):
        self.aggregates = aggregates
        self.groupby = groupby
        self.tuple_descriptor = None
        self.buckets = {}
        self.update_lock = RLock()
        kwargs = {windowsize[1]: int(windowsize[0])}
        self.windowsize = timedelta(**kwargs)
        self.window = None
        self.emptylist = []
    def update(self, updates):
        self.update_lock.acquire()
        output = self.emptylist
        for update in updates:
            if self.window == None:
                self.window = AggregateWindow(update.created_at, update.created_at + self.windowsize)
            # ignore all entries before the window
            test = self.window.windowtest(update.created_at)
            if test == AggregateWindowResult.AFTER:
                if output is self.emptylist:
                    output = []
                for bucket, aggs in self.buckets.items():
                    bucket.set_tuple_descriptor(self.tuple_descriptor)
                    for k,v in aggs.items():
                        setattr(bucket, k, v.value())
                    output.append(bucket)
                self.buckets = {}
                while self.window.windowtest(update.created_at) != AggregateWindowResult.IN:
                    self.window.advance(self.windowsize)
                test = AggregateWindowResult.IN
            if test == AggregateWindowResult.IN:
                bucket = update.generate_from_descriptor(self.groupby)
                if bucket not in self.buckets:
                    aggs = dict()
                    for aggregate in self.aggregates:
                        factory = aggregate.aggregate_factory
                        underlying = aggregate.underlying_fields
                        aggs[aggregate.alias] = factory(underlying)
                    self.buckets[bucket] = aggs
                for aggregate in self.buckets[bucket].values():
                    aggregate.update(update)
        self.update_lock.release()
        return output

class AggregateWindow():
    def __init__(self, start, end):
        self.start = start
        self.end = end
    def windowtest(self, timeval):
        if timeval < self.start:
            return AggregateWindowResult.BEFORE
        if timeval < self.end:
            return AggregateWindowResult.IN
        return AggregateWindowResult.AFTER
    def advance(self, windowsize):
        self.start = self.end
        self.end = self.end + windowsize

class AggregateWindowResult():
    BEFORE = 1
    IN = 2
    AFTER = 3

class Aggregate():
    def __init__(self, underlying_fields):
        self.underlying_fields = underlying_fields
        self.reset()
    def update(self, t):
        raise  NotImplementedError()
    def value(self):
        raise NotImplementedError()
    def reset(self):
        raise NotImplementedError()

def get_aggregate_factory(agg_func):
    if agg_func == QueryTokens.AVG:
        return Avg.create
    elif agg_func == QueryTokens.COUNT:
        return Count.create
    elif agg_func == QueryTokens.SUM:
        return Sum.create
    elif agg_func == QueryTokens.MIN:
        return Min.create
    elif agg_func == QueryTokens.MAX:
        return Max.create
    else:
        return None

class Avg(Aggregate):
    @classmethod
    def create(cls, underlying_fields):
        return Avg(underlying_fields)
    def update(self, t):
        self.sum += getattr(t, self.underlying_fields[0])
        self.count += 1
    def value(self):
        return self.sum/self.count
    def reset(self):
        self.sum = 0
        self.count = 0

class Count(Aggregate):
    @classmethod
    def create(cls, underlying_fields):
        return Count(underlying_fields)
    def update(self, t):
        self.count += 1
    def value(self):
        return self.count
    def reset(self):
        self.count = 0

class Sum(Aggregate):
    @classmethod
    def create(cls, underlying_fields):
        return Sum(underlying_fields)
    def update(self, t):
        self.sum += getattr(t, self.underlying_fields[0])
    def value(self):
        return self.sum
    def reset(self):
        self.sum = 0

class Min(Aggregate):
    @classmethod
    def create(cls, underlying_fields):
        return Min(underlying_fields)
    def update(self, t):
        val = getattr(t, self.underlying_fields[0])
        if (self.min is None) or (val < self.min):
            self.min = val
    def value(self):
        return self.min
    def reset(self):
        self.min = None

class Max(Aggregate):
    @classmethod
    def create(cls, underlying_fields):
        return Max(underlying_fields)
    def update(self, t):
        val = getattr(t, self.underlying_fields[0])
        if (self.max is None) or (val > self.max):
            self.max = val
    def value(self):
        return self.max
    def reset(self):
        self.max = None



from tweeql.exceptions import DbException
from tweeql.field_descriptor import ReturnType
from tweeql.exceptions import SettingsException
from tweeql.settings_loader import get_settings
from sqlalchemy import create_engine, Table, Column, Integer, Unicode, Float, DateTime, MetaData
from sqlalchemy.exc import ArgumentError, InterfaceError

settings = get_settings()

class StatusHandler(object):
    def __init__(self, batch_size):
        self.batch_size = batch_size
    def handle_statuses(self, statuses):
        raise NotImplementedError()
    def set_tuple_descriptor(self, descriptor):
        self.tuple_descriptor = descriptor

class PrintStatusHandler(StatusHandler):
    def __init__(self, batch_size, delimiter = u"|"):
        super(PrintStatusHandler, self).__init__(batch_size)
        self.delimiter = delimiter

    def handle_statuses(self, statuses):
        td = self.tuple_descriptor
        for status in statuses:
            vals = (unicode(val) for (alias, val) in status.as_iterable_visible_pairs())
            print self.delimiter.join(vals) + "\n"

class DbInsertStatusHandler(StatusHandler):
    engine = None

    def __init__(self, batch_size, tablename):
        super(DbInsertStatusHandler, self).__init__(batch_size)
        if DbInsertStatusHandler.engine == None:
            try:
                dburi = settings.DATABASE_URI
                dbconfig = None
                try:
                    dbconfig = settings.DATABASE_CONFIG
                except AttributeError:
                    pass
                if dbconfig == None:
                    DbInsertStatusHandler.engine = create_engine(dburi, echo=False, pool_size=1)
                else:
                    DbInsertStatusHandler.engine = create_engine(dburi, connect_args=dbconfig, echo=False, pool_size=1)
            except AttributeError, e:
                raise e
                raise SettingsException("To put results INTO a TABLE, please specify a DATABASE_URI in settings.py") 
            except ArgumentError, e:
                raise DbException(e)
        self.tablename = tablename

    def set_tuple_descriptor(self, descriptor):
        StatusHandler.set_tuple_descriptor(self, descriptor)
        metadata = MetaData()
        columns = []
        for alias in descriptor.aliases:
            desc = descriptor.get_descriptor(alias)
            if desc.visible:
                columns.append(self.db_col(alias, descriptor))
        columns.insert(0, Column('__id', Integer, primary_key=True))
        self.table = Table(self.tablename, metadata, *columns)
        try:
            metadata.create_all(bind=DbInsertStatusHandler.engine)
        except InterfaceError:
            raise SettingsException("Unable to connect to database.  Did you configure the connection properly?  Check DATABASE_URI and DATABASE_CONFIG in settings.py") 


        test = metadata.tables[self.tablename]
        self.verify_table()
   
    def db_col(self, alias, descriptor):
        return_type = descriptor.get_descriptor(alias).return_type
        type_val = None
        if return_type == ReturnType.INTEGER:
            type_val = Integer
        elif return_type == ReturnType.FLOAT:
            type_val = Float
        elif return_type == ReturnType.STRING:
            type_val = Unicode
        elif return_type == ReturnType.DATETIME:
            type_val = DateTime
        else:
            raise DbException("Unknown field return type: %s" % (return_type))
        return Column(alias, type_val)

    def verify_table(self):
        """
            Makes sure the table's schema is not different from the one in the database.
            This might happen if you try to load a query into a table which already
            exists and has a different schema.
        """
        metadata = MetaData()
        metadata.reflect(bind = DbInsertStatusHandler.engine)
        mine = str(self.table.columns)
        verified = str(metadata.tables[self.tablename].columns)
        if mine != verified:
            raise DbException("Table '%s' in the database has schema %s whereas the query's schema is %s" % (self.tablename, verified, mine)) 
 
    def handle_statuses(self, statuses):
        #from datetime import datetime
        #now = datetime.now()
        #print "handle called ", now 
        dicts = [dict(status.as_iterable_visible_pairs()) for status in statuses]
        conn = DbInsertStatusHandler.engine.connect()
        conn.execute(self.table.insert(), dicts)
        conn.close()
        #print "handle started ", now, " ended ", datetime.now()



from nltk.classify import NaiveBayesClassifier
from nltk.corpus import movie_reviews, stopwords
from nltk.collocations import BigramCollocationFinder
from nltk.metrics import BigramAssocMeasures
from nltk.probability import FreqDist, ConditionalFreqDist

import gzip
import os
import pickle
import re
import nltk

POSITIVE = "4"
NEGATIVE = "0"
NEUTRAL = "2"

mention_re = re.compile(ur"\@\S+", re.UNICODE)
url_re = re.compile(ur"((mailto\:|(news|(ht|f)tp(s?))\://){1}\S+)", re.UNICODE)
emoticon_re = re.compile(ur"\:\)|\:\-\)|\: \)|\:D|\=\)|\:\(\:\-\(\: \(", re.UNICODE)
tokenizer = nltk.RegexpTokenizer(r'\w+')

def word_feats(words):
    return dict([(word, True) for word in words])

def words_in_tweet(inputstr):
    outputstr = inputstr
    outputstr = mention_re.sub("MENTION_TOKEN", outputstr)
    outputstr = url_re.sub("URL_TOKEN", outputstr)
    outputstr = emoticon_re.sub("", outputstr)
    outputstr = outputstr.lower()
    return tokenizer.tokenize(outputstr)

# from http://groups.google.com/group/nltk-users/browse_thread/thread/be28ed12f87384ea
# Save Classifier 
def save_classifier(classifier): 
    fModel = open('BayesModel.pkl',"wb") 
    pickle.dump(classifier, fModel,1) 
    fModel.close() 
    os.system("rm BayesModel.pkl.gz") 
    os.system("gzip BayesModel.pkl") 
# Load Classifier
def load_classifier(): 
    os.system("gunzip BayesModel.pkl.gz") 
    fModel = open('BayesModel.pkl',"rb") 
    classifier = pickle.load(fModel) 
    fModel.close() 
    os.system("gzip BayesModel.pkl") 
    return classifier
# Package classifier for actual use
def package_classifier(to_pickle):
    fp = gzip.open('sentiment.pkl.gz', 'wb')
    pickle.dump(to_pickle, fp, 1)
    fp.close()



class StatusSource(object):
    TWITTER_FILTER = 1
    TWITTER_SAMPLE = 2

class QueryOperator(object):
    """
        QueryOperator represents an operator in the stream query plan.
        Anyone who implements this class should implement the filter method,
        which takes a list of updates on the stream and returns the ones
        which match the filter.
    """
    def __init__(self):
        pass
    def filter(self, updates, return_passes, return_fails):
        raise  NotImplementedError()
    def filter_params(self):
        """
            Returns a tuple with lists: (follow_ids, track_words)
        """
        raise  NotImplementedError()
    def can_query_stream(self):
        return False

class And(QueryOperator):
    """
        Logical AND between the operators that make up its children.
    """
    def __init__(self, children):
        QueryOperator.__init__(self)
        self.children = children
        self.can_query_stream_cache = self.can_query_stream_impl()
    def filter(self, updates, return_passes, return_fails):
        passes = updates
        fails = []
        for child in self.children:
            (passes, fails_local) = child.filter(passes, True, return_fails)
            if return_fails:
                fails.extend(fails_local)
        if not return_passes:
            passes = None
        if not return_fails:
            fails = None
        return (passes, fails)
    def filter_params(self):
        for child in self.children:
            if child.can_query_stream():
                return child.filter_params()
    def can_query_stream(self):
        return self.can_query_stream_cache
    def can_query_stream_impl(self):
        for child in self.children:
            if child.can_query_stream():
                return True
        return False

class Or(QueryOperator):
    """
        Logical OR between the operators that make up its children.
    """
    def __init__(self, children):
        QueryOperator.__init__(self)
        self.children = children
        self.can_query_stream_cache = self.can_query_stream_impl()
    def filter(self, updates, return_passes, return_fails):
        passes = []
        fails = updates
        for child in self.children:
            (passes_local, fails) = child.filter(fails, return_passes, True)
            if return_passes:
                passes.extend(passes_local)
        if not return_passes:
            passes = None
        if not return_fails:
            fails = None
        return (passes, fails)
    def filter_params(self):
        (follow_ids, track_words) = ([], []) 
        for child in self.children:
            if child.can_query_stream():
                (follow_local, track_local) = child.filter_params()
                follow_ids.extend(follow_local)
                track_words.extend(track_local)
        return (follow_ids, track_ids)
    def can_query_stream(self):
        return self.can_query_stream_cache
    def can_query_stream_impl(self):
        for child in self.children:
            if not child.can_query_stream():
                return False
        return True

class Not(QueryOperator):
    """
        Logical NOT on the child operator.
    """
    def __init__(self, child):
        QueryOperator.__init__(self)
        self.child = child
    def filter(self, updates, return_passes, return_fails):
        (passes, fails) = self.child.filter(updates, return_fails, return_passes)
        return (fails, passes)
    def filter_params(self):
        return child.filter_params()
    def can_query_stream(self):
        return child.can_query_stream()

class Follow(QueryOperator):
    """
        Passes updates which contain people in the follower list.
    """
    def __init__(self, ids):
        QueryOperator.__init__(self)
        self.ids = set(ids)
    def filter(self, updates, return_passes, return_fails):
        passes = [] if return_passes else None
        fails = [] if return_fails else None
        for update in updates:
            if update.author in self.ids:
                if return_passes:
                    passes.append(update.author)
            elif return_fails:
                fails.append(update.author)
        return (passes, fails)
    def filter_params(self):
        return (self.ids, None)
    def can_query_stream(self):
        return True 

class Contains(QueryOperator):
    """
        Passes updates which contain the desired term
    """
    def __init__(self, term):
        QueryOperator.__init__(self)
        self.term = term.lower()
    def filter(self, updates, return_passes, return_fails):
        passes = [] if return_passes else None
        fails = [] if return_fails else None
        for update in updates:
            if self.term in update.text.lower():
                if return_passes:
                    passes.append(update)
            elif return_fails:
                fails.append(update)
        return (passes, fails)
    def filter_params(self):
        return (None, [self.term])
    def can_query_stream(self):
        return True

class Location(QueryOperator):
    """
        Passes updates which are located in a geographic bounding box
    """
    def __init__(self, xmin, xmax, ymin, ymax):
        QueryOperator.__init__(self)
        (self.xmin, self.xmax, self.ymin, self.ymax) = (xmin, xmax, ymin, ymax)
    def filter(self, updates, return_passes, return_fails):
        passes = [] if return_passes else None
        fails = [] if return_fails else None
        for update in updates:
            x = update.geo.x
            y = update.geo.y
            if x >= xmin and x <= xmax and y >= ymin and y <= ymax:
                if return_passes:
                    passes.append(update)
            elif return_fails:
                fails.append(update)
        return (passes, fails)



# Started with http://pyparsing.wikispaces.com/file/view/simpleSQL.py 
# ( Copyright (c) 2003, Paul McGuire ) and extended from there
#
from pyparsing import Literal, CaselessLiteral, Word, Upcase, delimitedList, Optional, \
    Combine, Group, alphas, nums, alphanums, ParseException, Forward, oneOf, quotedString, \
    ZeroOrMore, restOfLine, Keyword

def gen_parser():
    # define SQL tokens
    selectStmt = Forward()
    selectToken = Keyword("select", caseless=True)
    fromToken   = Keyword("from", caseless=True)

    ident          = Word( alphas, alphanums + "_$" ).setName("identifier")
    columnName     = Upcase( delimitedList( ident, ".", combine=True ) )
    columnNameList = Group( delimitedList( columnName ) )
    tableName      = Upcase( delimitedList( ident, ".", combine=True ) )
    tableNameList  = Group( delimitedList( tableName ) )

    whereExpression = Forward()
    and_ = Keyword("and", caseless=True)
    or_ = Keyword("or", caseless=True)
    in_ = Keyword("in", caseless=True)

    E = CaselessLiteral("E")
    binop = oneOf("= != < > >= <= eq ne lt le gt ge contains", caseless=True)
    arithSign = Word("+-",exact=1)
    realNum = Combine( Optional(arithSign) + ( Word( nums ) + "." + Optional( Word(nums) )  |
                ( "." + Word(nums) ) ) + 
            Optional( E + Optional(arithSign) + Word(nums) ) )
    intNum = Combine( Optional(arithSign) + Word( nums ) + 
            Optional( E + Optional("+") + Word(nums) ) )

    columnRval = realNum | intNum | quotedString | columnName 
    whereCondition = Group(
            ( columnName + binop + columnRval ) |
            ( columnName + in_ + "(" + delimitedList( columnRval ) + ")" ) |
            ( columnName + in_ + "(" + selectStmt + ")" ) |
            ( "(" + whereExpression + ")" )
            )
    whereExpression << whereCondition + ZeroOrMore( ( and_ | or_ ) + whereExpression ) 

    # define the grammar
    selectStmt      << ( selectToken + 
            ( '*' | columnNameList ).setResultsName( "fields" ) + 
            fromToken + 
            tableNameList.setResultsName( "sources" ) + 
            Optional( Group( CaselessLiteral("where") + whereExpression ), "" ).setResultsName("where") )

    parser = selectStmt

    # define Oracle comment format, and ignore them
    oracleSqlComment = "--" + restOfLine
    parser.ignore( oracleSqlComment )
    
    return parser

def test( str ):
    print str,"->"
    parser = gen_parser()
    try:
        tokens = parser.parseString( str )
        print "tokens = ",        tokens
        print "tokens.fields =", tokens.fields
        print "tokens.sources =",  tokens.sources
        print "tokens.where =", tokens.where
    except ParseException, err:
        print " "*err.loc + "^\n" + err.msg
        print err
    print

def runtests():
    test( "SELECT * from XYZZY, ABC" )
    test( "select * from SYS.XYZZY" )
    test( "Select A from Sys.dual" )
    test( "Select A,B,C from Sys.dual" )
    test( "Select A, B, C from Sys.dual" )
    test( "Select A, B, C from Sys.dual, Table2   " )
    test( "Xelect A, B, C from Sys.dual" )
    test( "Select A, B, C frox Sys.dual" )
    test( "Select" )
    test( "Select &&& frox Sys.dual" )
    test( "Select A from Sys.dual where a in ('RED','GREEN','BLUE')" )
    test( "Select A from Sys.dual where a in ('RED','GREEN','BLUE') and b in (10,20,30)" )
    test( "Select A from Sys.dual where (a in ('RED','GREEN','BLUE') and b in (10,20,30)) OR (b in (10,20,30))" )
    test( "Select A,b from table1,table2 where table1.id eq table2.id -- test out comparison operators" )

if __name__ == '__main__':
    runtests()

"""
Test output:
>pythonw -u simpleSQL.py
SELECT * from XYZZY, ABC ->
tokens =  ['select', '*', 'from', ['XYZZY', 'ABC']]
tokens.fields = *
tokens.sources = ['XYZZY', 'ABC']

select * from SYS.XYZZY ->
tokens =  ['select', '*', 'from', ['SYS.XYZZY']]
tokens.fields = *
tokens.sources = ['SYS.XYZZY']

Select A from Sys.dual ->
tokens =  ['select', ['A'], 'from', ['SYS.DUAL']]
tokens.fields = ['A']
tokens.sources = ['SYS.DUAL']

Select A,B,C from Sys.dual ->
tokens =  ['select', ['A', 'B', 'C'], 'from', ['SYS.DUAL']]
tokens.fields = ['A', 'B', 'C']
tokens.sources = ['SYS.DUAL']

Select A, B, C from Sys.dual ->
tokens =  ['select', ['A', 'B', 'C'], 'from', ['SYS.DUAL']]
tokens.fields = ['A', 'B', 'C']
tokens.sources = ['SYS.DUAL']

Select A, B, C from Sys.dual, Table2    ->
tokens =  ['select', ['A', 'B', 'C'], 'from', ['SYS.DUAL', 'TABLE2']]
tokens.fields = ['A', 'B', 'C']
tokens.sources = ['SYS.DUAL', 'TABLE2']

Xelect A, B, C from Sys.dual ->
^
Expected 'select'
Expected 'select' (0), (1,1)

Select A, B, C frox Sys.dual ->
               ^
Expected 'from'
Expected 'from' (15), (1,16)

Select ->
      ^
Expected '*'
Expected '*' (6), (1,7)

Select &&& frox Sys.dual ->
       ^
Expected '*'
Expected '*' (7), (1,8)

>Exit code: 0
"""



from datetime import datetime
from field_descriptor import FieldDescriptor
from field_descriptor import FieldType
from field_descriptor import ReturnType
from tuple_descriptor import TupleDescriptor

def twitter_user_data_extractor(field):
    def factory():
        def extract(data):
            return getattr(data[TwitterFields.USER], field)
        return extract
    return factory

class TwitterFields:
    TEXT = "text"
    USER = "user"
    SSQL_USER_ID = "user_id"
    TWITTER_USER_ID = "id"
    SCREEN_NAME = "screen_name"
    LOCATION = "location"
    LANG = "lang"
    CREATED_AT = "created_at"
    PROFILE_IMAGE_URL = "profile_image_url"
    
# built here so we can refer to this field in the GroupBy operator. 
TwitterFields.created_field = FieldDescriptor(TwitterFields.CREATED_AT, [TwitterFields.CREATED_AT], FieldType.FIELD, ReturnType.DATETIME)

def twitter_tuple_descriptor():
    fields = [
        FieldDescriptor(TwitterFields.TEXT, [TwitterFields.TEXT], FieldType.FIELD, ReturnType.STRING),
        FieldDescriptor(TwitterFields.LOCATION, [], FieldType.FUNCTION, ReturnType.STRING, None, twitter_user_data_extractor(TwitterFields.LOCATION)),
        FieldDescriptor(TwitterFields.LANG, [], FieldType.FUNCTION, ReturnType.STRING, None, twitter_user_data_extractor(TwitterFields.LANG)),
        FieldDescriptor(TwitterFields.PROFILE_IMAGE_URL, [], FieldType.FUNCTION, ReturnType.STRING, None, twitter_user_data_extractor(TwitterFields.PROFILE_IMAGE_URL)),
        FieldDescriptor(TwitterFields.SSQL_USER_ID, [], FieldType.FUNCTION, ReturnType.INTEGER, None, twitter_user_data_extractor(TwitterFields.TWITTER_USER_ID)),
        FieldDescriptor(TwitterFields.SCREEN_NAME, [], FieldType.FUNCTION, ReturnType.STRING, None, twitter_user_data_extractor(TwitterFields.SCREEN_NAME)),
        TwitterFields.created_field,
    ]
    return TupleDescriptor(fields)




import copy
from field_descriptor import FieldDescriptor
from field_descriptor import FieldType
from field_descriptor import ReturnType
from tweeql.exceptions import QueryException

# fix deepcopy on instance methods
# see http://bugs.python.org/issue1515
import copy
import types

def _deepcopy_method(x, memo):
    return type(x)(x.im_func, copy.deepcopy(x.im_self, memo), x.im_class)
copy._deepcopy_dispatch[types.MethodType] = _deepcopy_method

class Tuple():
    def __init__(self):
        self.__tuple_descriptor = None
        self.__data = None
    def set_data(self, data):
        self.__data = data
    def set_tuple_descriptor(self, tuple_descriptor):
        self.__tuple_descriptor = tuple_descriptor
    def get_tuple_descriptor(self):
        return self.__tuple_descriptor
    def __getattr__(self, attr):
        field_descriptor = self.__tuple_descriptor.get_descriptor(attr)
        result = None
        if field_descriptor.field_type == FieldType.FUNCTION:
            uf = field_descriptor.underlying_fields
            func = field_descriptor.function
            args = [getattr(self, field) for field in uf]
            args.insert(0, self.__data)
            result = func(*args)
        elif field_descriptor.field_type == FieldType.LITERAL:
            result = field_descriptor.literal_value
        elif field_descriptor.underlying_fields[0] in self.__data:
            result = self.__data[field_descriptor.underlying_fields[0]]
        else:
            raise QueryException("Attribute not defined: %s" % (attr))
        if (field_descriptor.return_type == ReturnType.STRING) and isinstance(result, str):
            result = unicode(result)
        setattr(self, attr, result)
        return result
    def as_iterable_visible_pairs(self):
        for alias in self.__tuple_descriptor.aliases:
            if self.__tuple_descriptor.get_descriptor(alias).visible:
                yield (alias, getattr(self, alias))
    def generate_from_descriptor(self, tuple_descriptor):
        """
            Builds a new tuple from this one based on the tuple_descriptor that
            is passed in.
        """
        t = Tuple()
        d = {}
        t.set_data(d)
        for alias in tuple_descriptor.aliases:
            fields = self.__tuple_descriptor.get_descriptor(alias).underlying_fields
            for field in fields:
                setattr(t, field, getattr(self, field))
            setattr(t, alias, getattr(self, alias))
        t.set_tuple_descriptor(tuple_descriptor)
        return t
    def __hash__(self):
        if not self:
            return 0
        value = -1
        for alias in self.__tuple_descriptor.aliases:
            if self.__tuple_descriptor.get_descriptor(alias).visible:
                fieldval = hash(getattr(self, alias))
                if value == -1:
                    value = fieldval << 7
                else:
                    value = self.c_mul(1000003, value) ^ fieldval
        return value
    def __eq__(self, other):
        if isinstance(other, Tuple):
            if self.__tuple_descriptor is other.__tuple_descriptor: # strict object equality
                for alias in self.__tuple_descriptor.aliases:
                    if self.__tuple_descriptor.get_descriptor(alias).visible:
                        if getattr(self, alias) != getattr(other, alias):
                            return False
                return True
            else:
                return False
        else:
            return NotImplemented
    def c_mul(self, a, b):
        return eval(hex((long(a) * b) & 0xFFFFFFFFL)[:-1])

class TupleDescriptor():
    def __init__(self, field_descriptors = []):
        self.aliases = []
        self.descriptors = {}
        for descriptor in field_descriptors:
            self.add_descriptor(descriptor)
    def get_descriptor(self, alias):
        return self.descriptors.get(alias)
    def duplicate(self):
        return copy.deepcopy(self)
    def add_descriptor_list(self, descriptors):
        for descriptor in descriptors:
            self.add_descriptor(descriptor)
    def add_descriptor(self, descriptor):
        visible = descriptor.visible
        copy_descriptor = True
        if descriptor.alias in self.descriptors:
            if (self.descriptors[descriptor.alias].field_type != FieldType.UNDEFINED) and \
               (descriptor.field_type != FieldType.UNDEFINED) and \
               (self.descriptors[descriptor.alias] != descriptor):
                raise QueryException("The alias '%s' appears more than once in your query" % (descriptor.alias))
            # if one of the descriptors is visible, mark the stored one as
            # visible.
            visible = self.descriptors[descriptor.alias].visible or descriptor.visible
            if descriptor.field_type == FieldType.UNDEFINED:
                copy_descriptor = False
        else:
            self.aliases.append(descriptor.alias)
        if copy_descriptor:
            self.descriptors[descriptor.alias] = descriptor #copy.deepcopy(descriptor)
        self.descriptors[descriptor.alias].visible = visible



from itertools import chain
from pyparsing import ParseException
from tweeql import operators
from tweeql.aggregation import get_aggregate_factory
from tweeql.operators import StatusSource
from tweeql.exceptions import QueryException
from tweeql.exceptions import DbException
from tweeql.field_descriptor import FieldDescriptor
from tweeql.field_descriptor import FieldType
from tweeql.field_descriptor import ReturnType
from tweeql.function_registry import FunctionRegistry
from tweeql.query import Query
from tweeql.query import QueryTokens
from tweeql.tweeql_parser import gen_parser
from tweeql.status_handlers import PrintStatusHandler
from tweeql.status_handlers import DbInsertStatusHandler
from tweeql.tuple_descriptor import TupleDescriptor
from tweeql.twitter_fields import twitter_tuple_descriptor

def gen_query_builder():
    return QueryBuilder()

class QueryBuilder:
    """
        Generates a query from a declarative specification in a SQL-like syntax.
        This class is not thread-safe
    """
    def __init__(self):
        self.parser = gen_parser()
        self.function_registry = FunctionRegistry()
        self.unnamed_operator_counter = 0
        self.twitter_td = twitter_tuple_descriptor()
    def build(self, query_str):
        """
            Takes a Unicode string query_str, and outputs a query tree
        """
        try:
            parsed = self.parser.parseString(query_str)
        except ParseException,e:
            raise QueryException(e)

        source = self.__get_source(parsed)
        tree = self.__get_tree(parsed)
        handler = self.__get_handler(parsed)
        query = Query(tree, source, handler)
        return query
    def __get_source(self, parsed):
        source = parsed.sources[0]
        if source == QueryTokens.TWITTER:
            return StatusSource.TWITTER_FILTER
        elif source.startswith(QueryTokens.TWITTER_SAMPLE):
            return StatusSource.TWITTER_SAMPLE
        else:
            raise QueryException('Unknown query source: %s' % (source))
    def __get_tree(self, parsed):
        select = parsed.select.asList()[1:][0]
        where_clause = parsed.where.asList()
        groupby = parsed.groupby.asList()
        window = parsed.window.asList()
        window = None if window == [''] else window[1:]
        (tree, where_fields) = self.__parse_where(where_clause)
        tree = self.__add_select_and_aggregate(select, groupby, where_fields, window, tree)
        return tree
    def __get_handler(self, parsed):
        into = parsed.into.asList()
        handler = None
        if (into == ['']) or (into[1] == QueryTokens.STDOUT):
            handler = PrintStatusHandler(1)
        elif (len(into) == 3) and (into[1] == QueryTokens.TABLE):
            handler = DbInsertStatusHandler(1000, into[2])
        elif (len(into) == 3) and (into[1] == QueryTokens.STREAM):
            raise DbException("Putting results into a STREAM is not yet supported")
        else:
            raise QueryException("Invalid INTO clause")
        return handler
    def __parse_where(self, where_clause):
        tree = None
        where_fields = []
        if where_clause == ['']: # no where predicates
            tree = operators.AllowAll() 
        else:
            tree = self.__parse_clauses(where_clause[0][1:], where_fields)
        return (tree, where_fields)
    def __parse_clauses(self, clauses, where_fields):
        """
            Parses the WHERE clauses in the query.  Adds any fields it discovers to where_fields
        """
        self.__clean_list(clauses)
        if type(clauses) != list: # This is a token, not an expression 
            return clauses
        elif clauses[0] == QueryTokens.WHERE_CONDITION: # This is an operator expression
            return self.__parse_operator(clauses[1:], where_fields)
        else: # This is a combination of expressions w/ AND/OR
            # ands take precedent over ors, so 
            # A and B or C and D -> (A and B) or (C and D)
            ands = []
            ors = []
            i = 0
            while i < len(clauses):
                ands.append(self.__parse_clauses(clauses[i], where_fields))
                if i+1 == len(clauses):
                    ors.append(self.__and_or_single(ands))
                else:
                    if clauses[i+1] == QueryTokens.OR:
                        ors.append(self.__and_or_single(ands))
                        ands = []
                    elif clauses[i+1] == QueryTokens.AND:
                        pass
                i += 2
            # TODO: rewrite __and_or_single to handle the ors below just
            # like it does the ands above 
            if len(ors) == 1:
                return ors[0]
            else:
                return operators.Or(ors)
    def __parse_operator(self, clause, where_fields):
        if len(clause) == 3 and clause[1] == QueryTokens.CONTAINS:
            alias = self.__where_field(clause[0], where_fields)
            return operators.Contains(alias, self.__parse_rval(clause[2], allow_null=False))
        elif len(clause) == 3 and ((clause[1] == QueryTokens.EQUALS) or (clause[1] == QueryTokens.DOUBLE_EQUALS)):
            alias = self.__where_field(clause[0], where_fields)
            return operators.Equals(alias, self.__parse_rval(clause[2], allow_null=True))
        elif len(clause) == 3 and clause[1] == QueryTokens.EXCLAIM_EQUALS:
            alias = self.__where_field(clause[0], where_fields)
            return operators.Not(operators.Equals(alias, self.__parse_rval(clause[2], allow_null=True)))
    def __parse_rval(self, val, allow_null):
        if val == QueryTokens.NULL_TOKEN:
            if allow_null:
                return None
            else:
                raise QueryException("NULL appears in clause where it should not.")
        else:
            return val
    def __where_field(self, field, where_fields):
        (field_descriptors, verify) = self.__parse_field(field, self.twitter_td, False, False)
        alias = field_descriptors[0].alias
        # name the field whatever alias __parse_field gave it so it can be
        # passed to __parse_field in the future and have a consistent name
        if not ((len(field) >= 4) and (field[-2] == QueryTokens.AS)):
            field.append(QueryTokens.AS)
            field.append(alias)
        where_fields.append(field)
        return alias
    def __clean_list(self, list):
        self.__remove_all(list, QueryTokens.LPAREN)
        self.__remove_all(list, QueryTokens.RPAREN)

    def __remove_all(self, list, token):
        while token in list:
            list.remove(token)
    def __and_or_single(self, ands):
        if len(ands) == 1:
            return ands[0]
        else:
            return operators.And(ands)
    def __add_select_and_aggregate(self, select, groupby, where, window, tree):
        """
            select, groupby, and where are a list of unparsed fields
            in those respective clauses
        """
        tuple_descriptor = TupleDescriptor()
        fields_to_verify = []
        all_fields = chain(select, where)
        if groupby != ['']:
            groupby = groupby[1:][0]
            all_fields = chain(all_fields, groupby)
        self.__remove_all(groupby, QueryTokens.EMPTY_STRING)     
        for field in all_fields:
            (field_descriptors, verify) = self.__parse_field(field, self.twitter_td, True, False)
            fields_to_verify.extend(verify)
            tuple_descriptor.add_descriptor_list(field_descriptors)
        for field in fields_to_verify:
            self.__verify_and_fix_field(field, tuple_descriptor)
        
        # at this point, tuple_descriptor should contain a tuple descriptor
        # with fields/aliases that are correct (we would have gotten an
        # exception otherwise.  built select_descriptor/group_descriptor
        # from it
        select_descriptor = TupleDescriptor()
        group_descriptor = TupleDescriptor()
        aggregates = []
        for field in select:
            (field_descriptors, verify) = self.__parse_field(field, tuple_descriptor, True, True)
            select_descriptor.add_descriptor_list(field_descriptors)
            if field_descriptors[0].field_type == FieldType.AGGREGATE:
                aggregates.append(field_descriptors[0])
        # add WHERE clause fields as invisible attributes
        for field in where:
            (field_descriptors, verify) = self.__parse_field(field, tuple_descriptor, True, False)
            select_descriptor.add_descriptor_list(field_descriptors)
        if len(aggregates) > 0:
            if window == None:
                raise QueryException("Aggregate expression provided with no WINDOW parameter")
            for field in groupby:
                (field_descriptors, verify) = self.__parse_field(field, tuple_descriptor, True, True)
                group_descriptor.add_descriptor_list(field_descriptors)
            for alias in select_descriptor.aliases:
                select_field = select_descriptor.get_descriptor(alias)
                group_field = group_descriptor.get_descriptor(alias)
                if group_field == None and \
                   select_field.field_type != FieldType.AGGREGATE and \
                   select_field.visible:
                    raise QueryException("'%s' appears in the SELECT but is is neither an aggregate nor a GROUP BY field" % (alias))
            tree = operators.GroupBy(tree, group_descriptor, aggregates, window)
        tree.assign_descriptor(select_descriptor)
        return tree
    def __parse_field(self, field, tuple_descriptor, alias_on_complex_types, make_visible):
        """
            Returns a tuple containing (field_descriptors, fieldnames_to_verify)

            The first field in field_descriptors is the one requested to be parsed by this
            function call.  If the field turns out to be an aggregate or a user-defined
            function call, then field_descriptors will contain those parsed field descriptors
            as well, with their visible flag set to False.  

            fieldnames_to_verify is a list of field names that should be verified in order
            to ensure that at some point their alias is defined in an AS clause.
        """
        alias = None
        field_type = None
        return_type = None
        underlying_fields = None
        aggregate_factory = None
        literal_value = None
        func_factory = None
        fields_to_verify = []
        parsed_fds = []
        field_backup = list(field)
        self.__clean_list(field)
        
        # parse aliases if they exist
        if (len(field) >= 4) and (field[-2] == QueryTokens.AS):
            alias = field[-1]
            field = field[:-2]
        if (field[0] == QueryTokens.STRING_LITERAL) or \
           (field[0] == QueryTokens.INTEGER_LITERAL) or \
           (field[0] == QueryTokens.FLOAT_LITERAL): 
            alias = self.unnamed_operator_name()
            underlying_fields = []
            field_type = FieldType.LITERAL
            literal_value = field[1]
            if field[0] == QueryTokens.STRING_LITERAL:
                return_type = ReturnType.STRING
            elif field[0] == QueryTokens.INTEGER_LITERAL:
                return_type = ReturnType.INTEGER
                literal_value = int(literal_value)
            elif field[0] == QueryTokens.FLOAT_LITERAL:
                return_type = ReturnType.FLOAT
                literal_value = float(literal_value)
        elif field[0] == QueryTokens.COLUMN_NAME: # field or alias
            if alias == None:
                alias = field[1]
            field_descriptor = tuple_descriptor.get_descriptor(field[1])
            if field_descriptor == None: # underlying field not yet defined.  mark to check later
                field_type = FieldType.UNDEFINED
                underlying_fields = [field[1]]
                # check alias and underlying once this process is done to
                # find yet-undefined fields
                fields_to_verify.append(field[1])
                fields_to_verify.append(alias)
            else: # field found, copy information
                field_type = field_descriptor.field_type
                return_type = field_descriptor.return_type
                underlying_fields = field_descriptor.underlying_fields
                aggregate_factory = field_descriptor.aggregate_factory
                func_factory = field_descriptor.func_factory
        elif field[0] == QueryTokens.FUNCTION_OR_AGGREGATE: # function or aggregate  
            if alias == None:
                if alias_on_complex_types:
                    raise QueryException("Must specify alias (AS clause) for '%s'" % (field[1]))
                else:
                    alias = self.unnamed_operator_name()
            underlying_field_list = field[2:]
            underlying_fields = []
            for underlying in underlying_field_list:
                (parsed_fd_list, parsed_verify) = self.__parse_field(underlying, tuple_descriptor, False, False)
                for parsed_fd in parsed_fd_list:
                    parsed_fd.visible = False
                fields_to_verify.extend(parsed_verify)
                parsed_fds.extend(parsed_fd_list)
                underlying_fields.append(parsed_fd_list[0].alias)
            aggregate_factory = get_aggregate_factory(field[1])
            if aggregate_factory != None: # found an aggregate function
                field_type = FieldType.AGGREGATE
                return_type = ReturnType.FLOAT
            else:
                function_information = self.function_registry.get_function(field[1])
                if function_information != None:
                    field_type = FieldType.FUNCTION
                    func_factory = function_information.func_factory
                    return_type = function_information.return_type
                else:
                    raise QueryException("'%s' is neither an aggregate or a registered function" % (field[1]))
        else:
            raise QueryException("Empty field clause found: %s" % ("".join(field_backup)))
        fd = FieldDescriptor(alias, underlying_fields, field_type, return_type, aggregate_factory, func_factory, literal_value)
        fd.visible = make_visible
        parsed_fds.insert(0, fd)
        return (parsed_fds, fields_to_verify)
    
    def __verify_and_fix_field(self, field, tuple_descriptor):
        field_descriptor = tuple_descriptor.get_descriptor(field)
        error = False
        if field_descriptor == None:
            error = True
        elif field_descriptor.field_type == FieldType.UNDEFINED:
            if field == field_descriptor.underlying_fields[0]:
                error = True
            else:
                referenced_field_descriptor = \
                    self.__verify_and_fix_field(field_descriptor.underlying_fields[0],
                                                tuple_descriptor)
                field_descriptor.underlying_fields = referenced_field_descriptor.underlying_fields
                field_descriptor.field_type = referenced_field_descriptor.field_type
                field_descriptor.return_type = referenced_field_descriptor.return_type
                field_descriptor.aggregate_factory = referenced_field_descriptor.aggregate_factory
                field_descriptor.func_factory = referenced_field_descriptor.func_factory
                field_descriptor.function = referenced_field_descriptor.function
        if error:
            raise QueryException("Field '%s' is neither a builtin field nor an alias" % (field))
        else:
            return field_descriptor
    def unnamed_operator_name(self):
        self.unnamed_operator_counter += 1
        return "operand%d" % (self.unnamed_operator_counter)



#!/usr/bin/env python

from tweeql.exceptions import TweeQLException
from tweeql.query_runner import QueryRunner

import settings
import traceback

def main():
    runner = QueryRunner()
    try:
        while True:
            cmd = raw_input("tweeql> ");
            process_command(runner, cmd)
    except KeyboardInterrupt:
        print '\nGoodbye!'

def process_command(runner, cmd):
    try:
        runner.run_query(cmd, False)
    except KeyboardInterrupt:
        runner.stop_query()
    except TweeQLException, e:
        runner.stop_query()
        if settings.DEBUG:
            traceback.print_exc()
        else:
            print e

if __name__ == '__main__':
    main()



from getpass import getpass
from ssql.operators import StatusSource
from ssql.query_builder import gen_query_builder
from threading import RLock
from threading import Thread
from tweepy import StreamListener
from tweepy import Stream

class QueryRunner(StreamListener):
    def __init__(self, status_handler, batch_size = 10000):
        StreamListener.__init__(self)
        username = raw_input('Twitter username: ')
        password = getpass('Twitter password: ')
        self.status_lock = RLock()
        self.statuses = []
        self.batch_size = batch_size
        self.status_handler = status_handler
        self.query_builder = gen_query_builder()
        self.stream = Stream(username,
                             password,
                             self, # this object implements StreamListener
                             timeout = 600, # reconnect if no messages in 600s
                             retry_count = 20, # try reconnecting 20 times
                             retry_time = 10.0, # wait 10s if no HTTP 200
                             snooze_time = 1.0) # wait 1s if timeout in 600s
    def run_built_query(self, query_built):
        self.query = query_built
        # TODO: stop old query
        if self.query.source == StatusSource.TWITTER_FILTER:
            (follow_ids, track_words) = self.query.query_tree.filter_params()
            self.stream.filter(follow_ids, track_words, True)
        elif self.query.source == StatusSource.TWITTER_SAMPLE:
            self.stream.sample(None, True)
    def run_query(self, query_str):
        query_built = self.query_builder.build(query_str)
        self.run_built_query(query_built)
    def stop_query(self):
        self.stream.disconnect()
        self.flush_statuses()
    def filter_statuses(self, statuses, query, handler):
        (passes, fails) = query.query_tree.filter(statuses, True, False)
        handler.handle_statuses(passes)
    def flush_statuses(self):
        self.status_lock.acquire()
        filter_func = lambda: self.filter_statuses(self.statuses, self.query, self.status_handler)
        Thread(target = filter_func).start()
        self.statuses = []
        self.status_lock.release()

    """ StreamListener methods """
    def on_status(self, status):
        self.status_lock.acquire()
        self.statuses.append(status)
        if len(self.statuses) > self.batch_size:
            self.flush_statuses()
        self.status_lock.release()
    def on_error(self, status_code):
        print 'An error has occured! Status code = %s' % status_code
        return True # keep stream alive
    def on_timeout(self):
        print 'Snoozing Zzzzzz'

class PrintStatusHandler(object):
    def handle_statuses(self, statuses):
        for status in statuses:
            print status.text



# -*- coding: utf-8 -*-

from tweeql.field_descriptor import ReturnType
from tweeql.function_registry import FunctionInformation
from tweeql.function_registry import FunctionRegistry
from geopy import geocoders
from ordereddict import OrderedDict
from pkg_resources import resource_filename
from threading import RLock
from urllib2 import URLError

import gzip
import math
import re
import os
import pickle
import sys

class Temperature():
    fahr = re.compile(ur"(^| )(\-?\d+([.,]\d+)?)\s*\u00B0?(F$|F |Fahr)", re.UNICODE)
    celcius = re.compile(ur"(^| )(\-?\d+([.,]\d+)?)\s*\u00B0?(C$|C |Celsius)", re.UNICODE)
    return_type = ReturnType.FLOAT

    @staticmethod
    def factory():
        return Temperature().temperature_f

    def temperature_f(self, tuple_data, status):
        """ 
            Returns the temperature found in 'status' in Fahrenheit.  Captures
            both systems of temperature and then converts to Fahrenheit.
        """
        fahr_search = Temperature.fahr.search(status)
        temperature = None
        try:
            if fahr_search != None:
                temperature = fahr_search.group(2).replace(",", ".")
                temperature = float(temperature)
            else:
                celcius_search = Temperature.celcius.search(status)
                if celcius_search != None:
                    temperature = celcius_search.group(2).replace(",", ".")
                    temperature = float(temperature)
                    temperature = ((9.0/5) * temperature) + 32
        except ValueError:
            print "Encoding error on '%s'" % (status)
        return temperature

class Sentiment:
    classifier = None
    classinfo = None
    return_type = ReturnType.FLOAT
    constructor_lock = RLock()
    
    @staticmethod
    def factory():
        Sentiment.constructor_lock.acquire()
        if Sentiment.classifier == None:
            # Only import analysis if we have to: this means people who
            # don't use sentiment analysis don't have to install nltk.
            import tweeql.extras.sentiment
            import tweeql.extras.sentiment.analysis
            Sentiment.analysis = tweeql.extras.sentiment.analysis
            fname = resource_filename(tweeql.extras.sentiment.__name__, 'sentiment.pkl.gz')
            fp = gzip.open(fname)
            classifier_dict = pickle.load(fp)
            fp.close()
            Sentiment.classifier = classifier_dict['classifier']
            Sentiment.classinfo = { classifier_dict['pos_label'] :
                                      { 'cutoff': classifier_dict['pos_cutoff'],
                                        'value' : 1.0/classifier_dict['pos_recall'] },
                                    classifier_dict['neg_label'] :
                                      { 'cutoff': classifier_dict['neg_cutoff'],
                                        'value': -1.0/classifier_dict['neg_recall'] }
                                  }
        Sentiment.constructor_lock.release()
        return Sentiment().sentiment

    def sentiment(self, tuple_data, text):
        words = Sentiment.analysis.words_in_tweet(text)
        features = Sentiment.analysis.word_feats(words)
        dist = Sentiment.classifier.prob_classify(features)
        retval = 0
        maxlabel = dist.max()
        classinfo = Sentiment.classinfo[maxlabel]
        if dist.prob(maxlabel) > classinfo['cutoff']:
            retval = classinfo['value']
        return retval

class StringLength():
    return_type = ReturnType.INTEGER

    @staticmethod
    def factory():
        return StringLength().strlen

    def strlen(self, tuple_data, val):
        """ 
            Returns the length of val, which is a string
        """
        return len(val)
         
class Rounding():
    return_type = ReturnType.FLOAT

    @staticmethod
    def factory():
        return Rounding().floor

    def floor(self, tuple_data, val, nearest = 1):
        """ 
            Returns the largest integer multiple of 'nearest' that is less than or equal to val.
            If nearest is less than 1, you may see funny results because of floating 
            point arithmetic.
        """
        retval = val - (val % nearest) if val != None else None
        return retval

class Location:
    class LruDict(OrderedDict):
        def __setitem__(self, key, value):
            self.pop(key, None)
            OrderedDict.__setitem__(self, key, value)
        def compact_to_size(self, size):
            while len(self) > size:
                self.popitem(last=False)

    gn = geocoders.GeoNames()
    return_type = ReturnType.FLOAT
    LATLNG = "__LATLNG"
    LAT = "lat"
    LNG = "lng"
    cache = LruDict()
    cache_lock = RLock()

    @staticmethod
    def factory():
        return Location().get_latlng

    def get_latlng(self, tuple_data, lat_or_long):
        if not Location.LATLNG in tuple_data:
            tuple_data[Location.LATLNG] = Location.extract_latlng(tuple_data)
        val = None
        if tuple_data[Location.LATLNG] != None:
            if lat_or_long == Location.LAT:
                val = tuple_data[Location.LATLNG][0]
            elif lat_or_long == Location.LNG:
                val = tuple_data[Location.LATLNG][1]
        return val
    
    @staticmethod
    def extract_latlng(tuple_data):
        latlng = None
        if tuple_data["coordinates"] != None:
            coords = tuple_data["coordinates"]["coordinates"]
            latlng = (coords[1], coords[0])
        if latlng == None:
            loc = tuple_data["user"].location
            if (loc != None) and (loc != ""):
                loc = loc.lower()
                Location.cache_lock.acquire()
                latlng = Location.cache.get(loc, None)
                Location.cache_lock.release()
                if latlng == None:
                    latlng = Location.geonames_latlng(loc)
                Location.cache_lock.acquire()
                Location.cache[loc] = latlng
                Location.cache.compact_to_size(10000)
                Location.cache_lock.release()
        return latlng

    @staticmethod
    def geonames_latlng(loc):
        latlng = None
        try:
            g = Location.gn.geocode(loc.encode('utf-8'), exactly_one=False)
            for place, (lat, lng) in g:
                latlng = (lat, lng)
                break
        except URLError:
            e = sys.exc_info()[1]
            print "Unable to connect to GeoNames: %s" % (e)
        return latlng

class MeanOutliers():
    return_type = ReturnType.FLOAT

    class MeanGroup():
        def __init__(self):
            self.n = 0
            self.ewma = 0.0 # exponentially weighted moving avgerage
            self.ewmmd = 0.0 # exponentially weighted moving mean deviation
        def update_and_calculate(self, value):
            """
                Returns the number of mean deviations from the EWMA if
                the number of values previously recorded is >= 5.  Otherwise,
                returns -1.

                Updates the EWMA and EWMMD after calculating how many median deviations
                away the result is.
            """
            retval = -1
            diff = abs(self.ewma - value)
            if self.n >= 5: # only calculate meandevs if collected > 5 data pts.
                if self.ewmmd > 0:
                    meandevs = diff/self.ewmmd
                else:
                    meandevs = diff/.00001
                retval = meandevs
            
            # update ewma/ewmmd
            self.n += 1
            if self.n > 1:
                if self.n > 2:
                    self.ewmmd = (.125*diff) + (.875*self.ewmmd)
                else:
                    self.ewmmd = diff
                self.ewma = (.125*value) + (.875*self.ewma)
            else:
                self.ewma = value
            return retval 

    @staticmethod
    def factory():
        return MeanOutliers().nummeandevs

    def __init__(self):
        self.groups = dict()

    def nummeandevs(self, tuple_data, value, *group):
        """ 
            Calculates how many mean deviations from the exponentially 
            weighted moving average value is, given the other values that have been 
            given for the elements of this group.

            Uses the method TCP utilizes to estimate the round trip time 
            and mean deviation time of a potentially congested packet:
            http://tools.ietf.org/html/rfc2988

            The return value will be greater than or equal to 0 
            if it represents the number of mean deviations the value was away
            from the exponentially weighted moving average.
            If it is less than 0, the group does not have enough data to
            calculate mean deviations.

            What is a good outlier value?  If your data is normally distributed, then
            I experimentally found that 1 mean deviation is .8 standard deviations.
            For normally distributed data, 68% of values are within 1 standard
            deviation, 95% of values are within 2, and nearly 100% are within 3.
            Thus, 68% of values are within 1.25 mean deviations, 95% are within
            2.5 mean deviations, and almost 100% are within 3.75 mean deviations.
            A good mean deviation cutoff for legitimate values is thus in the range
            2.5-3.75.
        """
        mean_group = self.groups.get(group, None)
        if mean_group == None:
            mean_group = MeanOutliers.MeanGroup()
            self.groups[group] = mean_group

        return mean_group.update_and_calculate(value)

def register_default_functions():
    fr = FunctionRegistry()
    fr.register("temperatureF", FunctionInformation(Temperature.factory, Temperature.return_type))
    fr.register("tweetLatLng", FunctionInformation(Location.factory, Location.return_type))
    fr.register("floor", FunctionInformation(Rounding.factory, Rounding.return_type))
    fr.register("strlen", FunctionInformation(StringLength.factory, StringLength.return_type))
    fr.register("meanDevs", FunctionInformation(MeanOutliers.factory, MeanOutliers.return_type))
    fr.register("sentiment", FunctionInformation(Sentiment.factory, Sentiment.return_type))




from tweeql.exceptions import QueryException

class FunctionInformation():
    def __init__(self, func_factory, return_type):
        self.func_factory = func_factory
        self.return_type = return_type

class FunctionRegistry():
    __shared_dict = dict()
    def __init__(self):
        self.__dict__ = self.__shared_dict
        if len(self.__shared_dict.keys()) == 0:
            self.__functions = dict()
    def register(self, alias, function_information):
        if alias in self.__functions:
            raise QueryException("'%s' has already been registered" % (alias))
        self.__functions[alias] = function_information
    def get_function(self, alias):
        if alias not in self.__functions:
            raise QueryException("'%s' is not a registered function" % (alias))
        return self.__functions[alias]



# Started with http://pyparsing.wikispaces.com/file/view/simpleSQL.py 
# ( Copyright (c) 2003, Paul McGuire ) and extended from there
#
from pyparsing import Literal, CaselessLiteral, Word, upcaseTokens, delimitedList, Optional, \
    Combine, Group, alphas, nums, alphanums, ParseException, Forward, oneOf, quotedString, \
    ZeroOrMore, restOfLine, Keyword, removeQuotes, downcaseTokens
from tweeql.query import QueryTokens

def gen_parser():
    # define SQL tokens
    selectStmt = Forward()
    selectToken = Keyword(QueryTokens.SELECT, caseless=True)
    fromToken   = Keyword(QueryTokens.FROM, caseless=True)
    intoToken   = Keyword(QueryTokens.INTO, caseless=True)
    groupByToken   = Keyword(QueryTokens.GROUPBY, caseless=True)
    windowToken   = Keyword(QueryTokens.WINDOW, caseless=True)
    asToken = Keyword(QueryTokens.AS, caseless=True).setParseAction(upcaseTokens)
    nullToken = Keyword(QueryTokens.NULL, caseless=False).setParseAction(replace(QueryTokens.NULL_TOKEN))
    
    # Math operators
    E = CaselessLiteral("E")
    binop = oneOf("= != < > >= <= == eq ne lt le gt ge %s" % (QueryTokens.CONTAINS), caseless=True).setParseAction(upcaseTokens)
    arithSign = Word("+-",exact=1)
    realNum = Combine( Optional(arithSign) + ( Word( nums ) + "." + Optional( Word(nums) )  |
                ( "." + Word(nums) ) ) + 
            Optional( E + Optional(arithSign) + Word(nums) ) )
    intNum = Combine( Optional(arithSign) + Word( nums ) + 
            Optional( E + Optional("+") + Word(nums) ) )

    ident          = Word( alphas, alphanums + "_$" ).setName("identifier")
    columnName     = delimitedList( ident, ".", combine=True )
    columnName.setParseAction(label(QueryTokens.COLUMN_NAME))
    aliasName     = delimitedList( ident, ".", combine=True )
    stringLiteral = Forward()
    stringLiteral << quotedString
    stringLiteral.setParseAction(label(QueryTokens.STRING_LITERAL))
    intLiteral = Forward()
    intLiteral << intNum
    intLiteral.setParseAction(label(QueryTokens.INTEGER_LITERAL))
    floatLiteral = Forward()
    floatLiteral << realNum
    floatLiteral.setParseAction(label(QueryTokens.FLOAT_LITERAL))
    columnExpression = Forward()
    columnFunction = Word(alphas, alphanums) + "(" + Optional(delimitedList(Group( floatLiteral ) | Group ( stringLiteral ) | Group( intLiteral ) | columnExpression)) + ")" 
    columnFunction.setParseAction(label(QueryTokens.FUNCTION_OR_AGGREGATE))
    columnExpression << Group ( (columnFunction | columnName) + Optional( asToken + aliasName ) )
    columnExpressionList = Group( delimitedList( columnExpression ) )
    tableName      = delimitedList( ident, ".", combine=True ).setParseAction(upcaseTokens)
    tableNameList  = Group( delimitedList( tableName ) )
    timeExpression = Word( nums ) + oneOf("seconds minutes hours days", caseless=True).setParseAction(downcaseTokens)
   
    stdoutToken = Keyword(QueryTokens.STDOUT, caseless=True).setParseAction(upcaseTokens)
    tableToken = Keyword(QueryTokens.TABLE, caseless=True).setParseAction(upcaseTokens)
    streamToken = Keyword(QueryTokens.STREAM, caseless=True).setParseAction(upcaseTokens)
    intoLocation = stdoutToken | ( tableToken + ident ) | ( streamToken + ident ) 

    whereExpression = Forward()
    and_ = Keyword(QueryTokens.AND, caseless=True).setParseAction(upcaseTokens)
    or_ = Keyword(QueryTokens.OR, caseless=True).setParseAction(upcaseTokens)
    in_ = Keyword(QueryTokens.IN, caseless=True).setParseAction(upcaseTokens)

    columnRval = realNum | intNum | nullToken | columnExpression | quotedString.setParseAction(removeQuotes)
    whereCondition = Group(
            ( columnExpression + binop + columnRval ).setParseAction(label(QueryTokens.WHERE_CONDITION)) |
            ( columnExpression + in_ + "(" + delimitedList( columnRval ) + ")" ).setParseAction(label(QueryTokens.WHERE_CONDITION)) |
            ( columnExpression + in_ + "(" + selectStmt + ")" ).setParseAction(label(QueryTokens.WHERE_CONDITION)) |
            ( "(" + whereExpression + ")" )
            )
    whereExpression << whereCondition + ZeroOrMore( ( and_ | or_ ) + whereExpression ) 

    # define the grammar
    selectStmt      << (  
            Group ( selectToken + columnExpressionList ).setResultsName( "select" ) + 
            fromToken + 
            tableNameList.setResultsName( "sources" ) + 
            Optional(intoToken + intoLocation, "").setResultsName("into") + 
            Optional( Group( CaselessLiteral(QueryTokens.WHERE) + whereExpression ), "" ).setResultsName("where") +
            Optional ( groupByToken + columnExpressionList, "").setResultsName("groupby") +
            Optional ( windowToken + timeExpression, "").setResultsName("window")
            )

    parser = selectStmt

    # define Oracle comment format, and ignore them
    oracleSqlComment = "--" + restOfLine
    parser.ignore( oracleSqlComment )
    
    return parser

def test( str ):
    print str,"->"
    parser = gen_parser()
    try:
        tokens = parser.parseString( str )
        print "tokens = ",        tokens
        print "tokens.fields =", tokens.fields
        print "tokens.sources =",  tokens.sources
        print "tokens.where =", tokens.where
    except ParseException, err:
        print " "*err.loc + "^\n" + err.msg
        print err
    print

def label(l):
    """
        Returns a parseaction which prepends the tokens with the label l
    """
    def action(string, loc, tokens):
        newlist = [l]
        newlist.extend(tokens)
        return newlist
    return action

def replace(l):
    """
        Returns a parseaction which replaces the tokens with the token l
    """
    def action(string, loc, tokens):
        return [l]
    return action

def runtests():
    test( "SELECT * from XYZZY, ABC" )
    test( "select * from SYS.XYZZY" )
    test( "Select A from Sys.dual" )
    test( "Select A,B,C from Sys.dual" )
    test( "Select A, B, C from Sys.dual" )
    test( "Select A, B, C from Sys.dual, Table2   " )
    test( "Xelect A, B, C from Sys.dual" )
    test( "Select A, B, C frox Sys.dual" )
    test( "Select" )
    test( "Select &&& frox Sys.dual" )
    test( "Select A from Sys.dual where a in ('RED','GREEN','BLUE')" )
    test( "Select A from Sys.dual where a in ('RED','GREEN','BLUE') and b in (10,20,30)" )
    test( "Select A from Sys.dual where (a in ('RED','GREEN','BLUE') and b in (10,20,30)) OR (b in (10,20,30))" )
    test( "Select A,b from table1,table2 where table1.id eq table2.id -- test out comparison operators" )

if __name__ == '__main__':
    runtests()

"""
Test output:
>pythonw -u simpleSQL.py
SELECT * from XYZZY, ABC ->
tokens =  ['select', '*', 'from', ['XYZZY', 'ABC']]
tokens.fields = *
tokens.sources = ['XYZZY', 'ABC']

select * from SYS.XYZZY ->
tokens =  ['select', '*', 'from', ['SYS.XYZZY']]
tokens.fields = *
tokens.sources = ['SYS.XYZZY']

Select A from Sys.dual ->
tokens =  ['select', ['A'], 'from', ['SYS.DUAL']]
tokens.fields = ['A']
tokens.sources = ['SYS.DUAL']

Select A,B,C from Sys.dual ->
tokens =  ['select', ['A', 'B', 'C'], 'from', ['SYS.DUAL']]
tokens.fields = ['A', 'B', 'C']
tokens.sources = ['SYS.DUAL']

Select A, B, C from Sys.dual ->
tokens =  ['select', ['A', 'B', 'C'], 'from', ['SYS.DUAL']]
tokens.fields = ['A', 'B', 'C']
tokens.sources = ['SYS.DUAL']

Select A, B, C from Sys.dual, Table2    ->
tokens =  ['select', ['A', 'B', 'C'], 'from', ['SYS.DUAL', 'TABLE2']]
tokens.fields = ['A', 'B', 'C']
tokens.sources = ['SYS.DUAL', 'TABLE2']

Xelect A, B, C from Sys.dual ->
^
Expected 'select'
Expected 'select' (0), (1,1)

Select A, B, C frox Sys.dual ->
               ^
Expected 'from'
Expected 'from' (15), (1,16)

Select ->
      ^
Expected '*'
Expected '*' (6), (1,7)

Select &&& frox Sys.dual ->
       ^
Expected '*'
Expected '*' (7), (1,8)

>Exit code: 0
"""



from analysis import load_classifier, package_classifier, POSITIVE, NEUTRAL, NEGATIVE

classifier = load_classifier()
package_classifier({"classifier": classifier,
                    "pos_cutoff": 0.84,
                    "neg_cutoff": 0.66,
                    "pos_label": POSITIVE,
                    "neut_label": NEUTRAL,
                    "neg_label": NEGATIVE,
                    "pos_recall": 0.462963,
                    "pos_precision": 0.793651,
                    "neg_recall": 0.653333,
                    "neg_precision": 0.790323,
                   })



from getpass import getpass
from tweeql.builtin_functions import register_default_functions
from tweeql.exceptions import QueryException
from tweeql.exceptions import DbException
from tweeql.operators import StatusSource
from tweeql.query_builder import gen_query_builder
from tweeql.settings_loader import get_settings
from tweeql.tuple_descriptor import Tuple
from threading import RLock
from threading import Thread
from tweepy import Stream
from tweepy import StreamListener
from tweepy.auth import BasicAuthHandler

import time

settings = get_settings()

class QueryRunner(StreamListener):
    def __init__(self):
        register_default_functions()
        StreamListener.__init__(self)
        try:
            self.username = settings.TWITTER_USERNAME
            self.password = settings.TWITTER_PASSWORD
        except AttributeError:
            print "TWITTER_USERNAME and TWITTER_PASSWORD not defined in settings.py"
            self.username = raw_input('Twitter username: ')
            self.password = getpass('Twitter password: ')
        self.status_lock = RLock()
        self.statuses = []
        self.query_builder = gen_query_builder()
        self.stream = None
    def build_stream(self):
        if self.stream != None:
            self.stop_query()
            time.sleep(.01) # make sure old stream has time to disconnect
        self.stream = Stream(BasicAuthHandler(self.username, self.password),
                             self, # this object implements StreamListener
                             timeout = 600, # reconnect if no messages in 600s
                             retry_count = 20, # try reconnecting 20 times
                             retry_time = 10.0, # wait 10s if no HTTP 200
                             snooze_time = 1.0) # wait 1s if timeout in 600s
    def run_built_query(self, query_built, async):
        self.build_stream()
        self.query = query_built
        self.query.handler.set_tuple_descriptor(self.query.get_tuple_descriptor())
        if self.query.source == StatusSource.TWITTER_FILTER:
            no_filter_exception = QueryException("You haven't specified any filters that can query Twitter.  Perhaps you want to query TWITTER_SAMPLE?")
            try:
                (follow_ids, track_words) = self.query.query_tree.filter_params()
                if (follow_ids == None) and (track_words == [None]):
                    raise no_filter_exception
                self.stream.filter(follow_ids, track_words, async)
            except NotImplementedError:
                raise no_filter_exception
        elif self.query.source == StatusSource.TWITTER_SAMPLE:
            self.stream.sample(None, async)
    def run_query(self, query_str, async):
        if isinstance(query_str, str):
            query_str = unicode(query_str, 'utf-8')
        query_built = self.query_builder.build(query_str)
        self.run_built_query(query_built, async)
    def stop_query(self):
        if self.stream != None:
            self.stream.disconnect()
            self.flush_statuses()
    def filter_statuses(self, statuses, query):
        (passes, fails) = query.query_tree.filter(statuses, True, False)
        query.handler.handle_statuses(passes)
    def flush_statuses(self):
        self.status_lock.acquire()
        if len(self.statuses) > 0:
            filter_func = lambda s=self.statuses, q=self.query: self.filter_statuses(s,q)
            t = Thread(target = filter_func)
            t.start()
            self.statuses = []
        self.status_lock.release()

    """ StreamListener methods """
    def on_status(self, status):
        self.status_lock.acquire()
        t = Tuple()
        t.set_tuple_descriptor(None)
        t.set_data(status.__dict__)
        self.statuses.append(t)
        if len(self.statuses) >= self.query.handler.batch_size:
            self.flush_statuses()
        self.status_lock.release()
    def on_error(self, status_code):
        print 'An error has occured! Status code = %s' % status_code
        return True # keep stream alive
    def on_timeout(self):
        print 'Snoozing Zzzzzz'
    def on_limit(self, limit_data):
        print "Twitter rate-limited this query.  Since query start, Twitter dropped %d messages." % (limit_data)



from analysis import load_classifier, words_in_tweet, POSITIVE, NEGATIVE, NEUTRAL, word_feats
import collections
import datetime
import nltk

def drange(start, stop, step):
    r = start
    while r <= stop:
        yield r
        r += step

testfile = open('testdata.manual.2009.05.25')

print "Loading classifier"
classifier = load_classifier()

print "Running test"
print "prob, pos prec, pos rec, neg prec, neg rec"
for prob in drange(0.5, 1.0, .01):
    right = 0
    wrong = 0
    refsets = collections.defaultdict(set)
    testsets = collections.defaultdict(set)
    testfile.seek(0)
    count = 0
    for line in testfile:
        parts = line.split(";;")
        dist = classifier.prob_classify(word_feats(words_in_tweet(parts[5])))
        if dist.prob(dist.max()) > prob:
            realguess = dist.max()
        else:
            realguess = NEUTRAL
        refsets[parts[0]].add(count)
        testsets[realguess].add(count)
        count += 1
    print "%f, %f, %f, %f, %f" % (prob, nltk.metrics.precision(refsets[POSITIVE], testsets[POSITIVE]), nltk.metrics.recall(refsets[POSITIVE],testsets[POSITIVE]), nltk.metrics.precision(refsets[NEGATIVE], testsets[NEGATIVE]), nltk.metrics.recall(refsets[NEGATIVE], testsets[NEGATIVE]))



class Query():
    def __init__(self, query_tree, source, handler):
        self.query_tree = query_tree
        self.source = source
        self.handler = handler
    def get_tuple_descriptor(self):
        return self.query_tree.get_tuple_descriptor()

class QueryTokens:
    SELECT = "SELECT"
    FROM = "FROM"
    WHERE = "WHERE"
    GROUPBY = "GROUP BY"
    INTO = "INTO"
    WINDOW = "WINDOW"
    TWITTER = "TWITTER"
    TWITTER_SAMPLE = "TWITTER_SAMPLE"
    LPAREN = "("
    RPAREN = ")"
    AND = "AND"
    OR = "OR"
    CONTAINS = "CONTAINS"
    AS = "AS"
    IN = "IN"
    AVG = "AVG"
    COUNT = "COUNT"
    SUM = "SUM"
    MIN = "MIN"
    MAX = "MAX"
    STDOUT = "STDOUT"
    TABLE = "TABLE"
    STREAM = "STREAM"
    EQUALS = "="
    DOUBLE_EQUALS = "=="
    EXCLAIM_EQUALS = "!="
    NULL = "NULL"
    EMPTY_STRING = ""
    STRING_LITERAL = "$$$STRING_LITERAL$$$"
    INTEGER_LITERAL = "$$$INTEGER_LITERAL$$$"
    FLOAT_LITERAL = "$$$FLOAT_LITERAL$$$"
    NULL_TOKEN = "$$$NULL_TOKEN$$$"
    WHERE_CONDITION = "$$$WHERE_CONDITION$$$"
    FUNCTION_OR_AGGREGATE = "$$$FUNCTION_OR_AGGREGATE$$$"
    COLUMN_NAME = "$$$COLUMN_NAME$$$"



import os
import sys

settings_cache = None

def get_settings():
    global settings_cache
    if settings_cache == None:
        sys.path.insert(0, os.getcwd())
        try:
            import settings
        except ImportError, e:
            print "It looks like you don't have a settings.py with basic project settings."
            sys.exit(-1)
        settings_cache = settings
        sys.path.pop(0)
    return settings_cache



from ssql import operators
from ssql.ssql_parser import gen_parser
from ssql.operators import StatusSource
from ssql.query import Query

class QueryTokens:
    TWITTER = "TWITTER"
    TWITTER_SAMPLE = "TWITTER_SAMPLE"
    LPAREN = "("
    RPAREN = ")"
    AND = "AND"
    OR = "OR"

def gen_query_builder():
    return QueryBuilder()

class QueryBuilder:
    def __init__(self):
        self.parser = gen_parser()
    def build(self, query_str):
        parsed = self.parser.parseString(query_str)
        source = self.__get_source(parsed)
        tree = self.__get_tree(parsed)
        query = Query(tree, source)
        return query
    def __get_source(self, parsed):
        source = parsed.sources[0]
        if source == QueryTokens.TWITTER:
            return StatusSource.TWITTER_FILTER
        elif source.startswith(QueryTokens.TWITTER_SAMPLE):
            return StatusSource.TWITTER_SAMPLE
    def __get_tree(self, parsed):
        return self.__parse_multi(parsed.where.asList()[0][1:])
    def __parse_multi(self, groups):
        self.__remove_all(groups, QueryTokens.LPAREN)
        self.__remove_all(groups, QueryTokens.RPAREN)
        if len(groups) == 1:
            return self.__parse_single(groups[0])
        else:
            # ands take precedent over ors, so 
            # A and B or C and D -> (A and B) or (C and D)
            ands = []
            ors = []
            i = 0
            while i < len(groups):
                ands.append(self.__parse_multi(groups[i]))
                # TODO: this can be simplified
                if i+1 == len(groups):
                    ors.append(self.__and_or_single(ands))
                else:
                    if groups[i+1] == QueryTokens.OR:
                        ors.append(self.__and_or_single(ands))
                        ands = []
                    elif groups[i+1] == QueryTokens.AND:
                        pass
                i += 2
            
            if len(ors) == 1:
                return ors[0]
            else:
                return operators.Or(ors)
    def __parse_single(self, clause):
        if len(clause) == 3 and clause[1] == "contains":
            return operators.Contains(clause[2].lstrip("'\"").rstrip("'\""))
    def __remove_all(self, list, token):
        while token in list:
            list.remove(token)
    def __and_or_single(self, ands):
        if len(ands) == 1:
            return ands[0]
        else:
            return operators.And(ands)




#!/usr/bin/env python
try:
    from setuptools import setup, find_packages
except ImportError:
    import ez_setup
    ez_setup.use_setuptools()
from setuptools import setup, find_packages
import tweeql

setup(name="tweeql",
      version=tweeql.__version__,
      description="A SQL-like query language for the Twitter API",
      license="BSD",
      author="Adam Marcus",
      author_email="tweeql@googlegroups.com",
      url="http://github.com/marcua/tweeql",
      packages = find_packages(),
      scripts = ['tweeql/bin/tweeql-command-line.py'],
      package_data = { 'tweeql' : ['extras/sentiment/sentiment.pkl.gz'] },
      install_requires = ['pyparsing>=1.5.2', 'sqlalchemy>=0.6.1', 'ordereddict', 'geopy', 'tweepy>=1.8'],
      keywords= "twitter query sql library",
      zip_safe = True)



class TweeQLException(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class QueryException(TweeQLException):
    pass

class DbException(TweeQLException):
    pass

class SettingsException(TweeQLException):
    pass



from analysis import save_classifier, words_in_tweet, POSITIVE, NEGATIVE, NEUTRAL
from nltk.metrics.association import BigramAssocMeasures

import collections, itertools
import datetime
import nltk
import nltk.classify.util, nltk.metrics
import os
import pickle
import re


# lots of stuff from http://streamhacker.com/2010/06/16/text-classification-sentiment-analysis-eliminate-low-information-features/
# and http://nltk.googlecode.com/svn/trunk/doc/book/ch06.html

#smilefile = open('smiley.txt.processed.2009.05.25')
#frownfile = open('frowny.txt.processed.2009.05.25')
smilefile = open('happy.txt')
frownfile = open('sad.txt')

def features(feat_func, handle, label):
    print "Generating features for '%s'" % (label)
    print datetime.datetime.now()
    return [((feat_func(words_in_tweet(line))), label) for line in handle]

def update_wordcount(word_fd, label_word_fd, handle, label):
    print "Counting '%s'" % (label)
    print datetime.datetime.now()
    for line in handle:
        for word in words_in_tweet(line):
            word_fd.inc(word)
            label_word_fd[label].inc(word)
    handle.seek(0)

word_fd = nltk.probability.FreqDist()
label_word_fd = nltk.probability.ConditionalFreqDist()
update_wordcount(word_fd, label_word_fd, smilefile, POSITIVE)
update_wordcount(word_fd, label_word_fd, frownfile, NEGATIVE)

pos_word_count = label_word_fd[POSITIVE].N()
neg_word_count = label_word_fd[POSITIVE].N()
total_word_count = pos_word_count + neg_word_count

print "Finding top words"
word_scores = {}
for word, freq in word_fd.iteritems():
    pos_score = BigramAssocMeasures.chi_sq(label_word_fd[POSITIVE][word],
        (freq, pos_word_count), total_word_count)
    neg_score = BigramAssocMeasures.chi_sq(label_word_fd[NEGATIVE][word],
        (freq, neg_word_count), total_word_count)
    word_scores[word] = pos_score + neg_score

best = sorted(word_scores.iteritems(), key=lambda (w,s): s, reverse=True)[:10000]
bestwords = set([w for w, s in best])
print "Best words"
#print bestwords
def best_word_feats(words):
    return dict([(word, True) for word in words if word in bestwords])

posfeats = features(best_word_feats, smilefile, POSITIVE)
negfeats = features(best_word_feats, frownfile, NEGATIVE)
classifier = nltk.NaiveBayesClassifier.train(posfeats + negfeats)
save_classifier(classifier)
